from lionfuncs import to_dict

from lion_core.protocols.models import FieldModel

_description = (
    "Specify the name of the function to execute. **Choose "
    "from the provided `tool_schemas`; do not invent function names.**"
    "Only provide function names if tool_schemas are provided. Otherwise, "
    "must leave blank or set to None."
)


def validate_args(cls, value):
    return to_dict(value, fuzzy_parse=True)


FUNCTION_FIELD = FieldModel(
    name="function",
    default=None,
    annotation=str | None,
    title="Function",
    description=_description,
    examples=["add", "multiply", "divide"],
    validator=validate_args,
)


__all__ = ["FUNCTION_FIELD"]



import inspect

from lionfuncs import copy, validate_boolean
from pydantic import (
    BaseModel,
    Field,
    PrivateAttr,
    create_model,
    field_validator,
    model_validator,
)
from pydantic.fields import FieldInfo
from typing_extensions import Self

from lion_core.protocols.fields.action_responses_ import ACTION_RESPONSES_FIELD
from lion_core.protocols.models import FieldModel, ParamsModel


class PrepFieldParamsModel(ParamsModel):

    parameters: dict[str, FieldInfo] = Field(default_factory=dict)
    base: type[BaseModel] = Field(default=BaseModel)
    exclude_fields: list = Field(default_factory=list)
    include_fields: list = Field(default_factory=list)
    field_descriptions: dict = Field(default_factory=dict)
    use_base_kwargs: bool = Field(default=False)
    use_all_fields: bool = Field(default=False)
    name: str = Field(default="StepModel")
    class_kwargs: dict = Field(default_factory=dict)
    field_models: list[FieldModel] = Field(default_factory=list)
    request_model_type: type[BaseModel] | None = Field(default=None)
    response_model_type: type[BaseModel] | None = Field(default=None)
    request_model: BaseModel | None = Field(default=None)
    response_model: BaseModel | None = Field(default=None)
    inherit_base: bool = Field(default=True)
    str_dict_response: str | dict | None = None
    use_fields: dict[str, tuple] | None = None

    _validators: dict[str, callable] | None = PrivateAttr(default=None)

    @field_validator("parameters", mode="before")
    def validate_parameters(cls, value):
        if value is None:
            return {}
        if not isinstance(value, dict):
            raise ValueError("Fields must be a dictionary.")
        for k, v in value.items():
            if not isinstance(k, str):
                raise ValueError("Field names must be strings.")
            if not isinstance(v, FieldInfo):
                raise ValueError("Field values must be FieldInfo objects.")
        return copy(value)

    @field_validator("base", mode="before")
    def validate_base(cls, value):
        if value is None:
            return BaseModel
        if isinstance(value, type) and issubclass(value, BaseModel):
            return value
        if isinstance(value, BaseModel):
            return value.__class__
        raise ValueError("Base must be a BaseModel subclass or instance.")

    @field_validator("exclude_fields", "include_fields", mode="before")
    def validate_fields(cls, value):
        if value is None:
            return []
        if isinstance(value, dict):
            value = list(value.keys())
        if isinstance(value, list):
            value = list(value)
        if isinstance(value, list):
            if not all(isinstance(i, str) for i in value):
                raise ValueError("Field names must be strings.")
            return copy(value)
        raise ValueError("Fields must be a list, set, or dictionary.")

    @field_validator("field_descriptions", mode="before")
    def validate_field_descriptions(cls, value):
        if value is None:
            return {}
        if not isinstance(value, dict):
            raise ValueError("Field descriptions must be a dictionary.")
        for k, v in value.items():
            if not isinstance(k, str):
                raise ValueError("Field names must be strings.")
            if not isinstance(v, str):
                raise ValueError("Field descriptions must be strings.")
        return value

    @field_validator("use_base_kwargs", "use_all_fields", mode="before")
    def validate_use_base_kwargs(cls, value):
        try:
            return validate_boolean(value)
        except Exception as e:
            raise ValueError(
                f"Failed to convert {value} to a boolean. Error: {e}"
            ) from e

    @field_validator("name", mode="before")
    def validate_name(cls, value):
        if value is None:
            return "StepModel"
        if not isinstance(value, str):
            raise ValueError("Name must be a string.")
        return value

    @field_validator("class_kwargs", mode="before")
    def validate_class_kwargs(cls, value):
        if value is None:
            return {}
        if not isinstance(value, dict):
            raise ValueError("Class kwargs must be a dictionary.")
        for k, v in value.items():
            if not isinstance(k, str):
                raise ValueError("Class keyword names must be strings.")
        return value

    @field_validator("field_models", mode="before")
    def validate_field_models(cls, value):
        if value is None:
            return []
        value = [value] if not isinstance(value, list) else value
        if not all(isinstance(i, FieldModel) for i in value):
            raise ValueError("Field models must be FieldModel objects.")
        return value

    @model_validator(mode="after")
    def validate_param_model(self) -> Self:
        if any([i in self.exclude_fields for i in self.include_fields]):
            raise ValueError(
                "Operation include fields and exclude fields cannot"
                " have common elements."
            )
        if self.field_models:
            self.parameters.update(
                {f.name: f.field_info for f in self.field_models}
            )

        if self.base:
            if self.use_all_fields and hasattr(self.base, "all_fields"):
                self.parameters.update(copy(self.base.all_fields))
            elif hasattr(self.base, "model_fields"):
                self.parameters.update(copy(self.base.model_fields))

        use_keys = list(self.parameters.keys())

        if self.exclude_fields:
            use_keys = [i for i in use_keys if i not in self.exclude_fields]

        self.parameters = {
            k: v for k, v in self.parameters.items() if k in use_keys
        }

        validators = {}
        for i in self.field_models:
            if i.field_validator is not None:
                validators.update(i.field_validator)
        self._validators = validators

        if self.field_descriptions:
            for field_name, description in self.field_descriptions.items():
                if description and field_name in self.parameters:
                    self.parameters[field_name] = self.parameters[
                        field_name
                    ].model_copy(update={"description": description})

        # Prepare class attributes
        class_kwargs = {}
        if self.use_base_kwargs:
            class_kwargs.update(
                {
                    k: getattr(self.base, k)
                    for k in self.base.__dict__
                    if not k.startswith("__")
                }
            )

        if hasattr(self.base, "class_name"):
            if callable(self.base.class_name):
                self.name = self.base.class_name()
            else:
                self.name = self.base.class_name
        elif inspect.isclass(self.base):
            self.name = self.base.__name__

            self.use_fields = {
                k: (v.annotation, v) for k, v in self.parameters.items()
            }
        return self

    @classmethod
    def as_request_model_type(cls, **kwargs) -> "PrepFieldParamsModel":
        self = cls.model_validate(kwargs)

        if not self.request_model_type:
            self.request_model_type = create_model(
                self.name + "Request",
                __config__=kwargs.pop("config_dict", None),
                __doc__=kwargs.pop("doc", None),
                __base__=self.base if self.inherit_base else BaseModel,
                __validators__=self._validators,
                __cls_kwargs__=self.class_kwargs,
                **self.use_fields,
            )
            if kwargs.pop("frozen", False) is True:
                self.request_model_type.model_config.frozen = True
        return self

    @classmethod
    def as_response_model_type(
        cls,
        request_params: "PrepFieldParamsModel",
        field_models: list[FieldModel] = None,
        **kwargs,
    ) -> "PrepFieldParamsModel":
        if (
            request_params.request_model is None
            or request_params.str_dict_response is None
        ):
            raise ValueError(
                "A completed Step Request model must be provided,"
                " you cannot generate a response model without a "
                "request model."
            )
        params = request_params.request_model_type.model_fields
        field_models = field_models or []
        field_models.extend(request_params.field_models)

        self = cls(
            parameters=params,
            base=request_params.base,
            exclude_fields=request_params.exclude_fields,
            use_base_kwargs=request_params.use_base_kwargs,
            name=request_params.name,
            inherit_base=request_params.inherit_base,
            field_descriptions=request_params.field_descriptions,
            use_all_fields=request_params.use_all_fields,
            field_models=field_models,
            validators={
                **(kwargs.pop("response_validators", None) or {}),
                **(request_params._validators or {}),
            },
        )

        self.response_model_type = create_model(
            self.name + "Response",
            __config__=kwargs.pop("response_config_dict", None),
            __doc__=kwargs.pop("response_doc", None),
            __base__=self.base if self.inherit_base else BaseModel,
            __validators__=self._validators,
            __cls_kwargs__=self.class_kwargs,
            **self.use_fields,
        )
        if kwargs.pop("frozen", False) is True:
            self.request_model_type.model_config.frozen = True

        return self

    @classmethod
    def parse_request_to_response_model(
        cls,
        request_params: "PrepFieldParamsModel",
        response_validators: dict = None,
        frozen_reponse: bool = False,
        response_config_dict=None,
        response_doc=None,
        action_responses: list = None,
        **kwargs,
    ) -> dict:

        if (
            action_responses
            and ACTION_RESPONSES_FIELD not in request_params.field_models
        ):
            request_params.field_models.append(ACTION_RESPONSES_FIELD)

        response_param = cls.as_response_model_type(
            request_params=request_params,
            response_validators=response_validators,
            frozen_reponse=frozen_reponse,
            response_config_dict=response_config_dict,
            response_doc=response_doc,
        )

        if request_params.request_model:
            kwargs.update(request_params.request_model.model_dump())
        elif isinstance(request_params.str_dict_response, dict):
            kwargs.update(request_params.str_dict_response)

        return response_param.response_model_type.model_validate(kwargs)



from __future__ import annotations

import re
from typing import Any

from lionfuncs import to_dict, to_json
from pydantic import BaseModel, Field, field_validator

from lion_core.protocols.fields.arguments_ import ARGUMENTS_FIELD
from lion_core.protocols.fields.function_ import FUNCTION_FIELD


class ActionRequestModel(BaseModel):

    function: str | None = FUNCTION_FIELD.field_info
    arguments: dict[str, Any] = ARGUMENTS_FIELD.field_info

    @field_validator("arguments", **ARGUMENTS_FIELD.validator_kwargs)
    def validate_arguments(cls, value: Any) -> dict[str, Any]:
        return FUNCTION_FIELD.validator(cls, value)

    @classmethod
    def create(cls, content: str) -> list[ActionRequestModel]:
        try:
            content = _parse_action_request(content)
            if content:
                return [cls.model_validate(i) for i in content]
            return []
        except Exception:
            return []


class ActionResponseModel(BaseModel):

    function: str = Field(default_factory=str)
    arguments: dict[str, Any] = Field(default_factory=dict)
    output: Any = None


def _parse_action_request(content: str | dict) -> list[dict]:

    json_blocks = []

    if isinstance(content, BaseModel):
        json_blocks = [content.model_dump()]

    elif content and isinstance(content, dict):
        json_blocks = [content]

    elif isinstance(content, str):
        json_blocks = to_json(content, fuzzy_parse=True)
        if not json_blocks:
            pattern2 = r"```python\s*(.*?)\s*```"
            _d = re.findall(pattern2, content, re.DOTALL)
            json_blocks = [to_dict(match, fuzzy_parse=True) for match in _d]
            json_blocks = [i for i in json_blocks if i]

    out = []

    for i in json_blocks:
        j = {}
        if isinstance(i, dict):
            for k, v in i.items():
                k = (
                    k.replace("action_", "")
                    .replace("recipient_", "")
                    .replace("s", "")
                )
                if k in ["name", "function", "recipient"]:
                    j["function"] = v
                elif k in ["parameter", "argument", "arg"]:
                    j["arguments"] = to_dict(
                        v, str_type="json", fuzzy_parse=True, suppress=True
                    )
            if (
                j
                and all(key in j for key in ["function", "arguments"])
                and j["arguments"]
            ):
                out.append(j)

    return out



from .adapters.adapter import Adapter, AdapterRegistry
from .models import FieldModel, OperableModel, SchemaModel
from .operatives.action_model import ActionRequestModel, ActionResponseModel
from .operatives.reason_model import ReasonModel
from .operatives.step import Step
from .registries._component_registry import ComponentAdapterRegistry
from .registries._pile_registry import PileAdapterRegistry

__all__ = [
    "Adapter",
    "AdapterRegistry",
    "ComponentAdapterRegistry",
    "PileAdapterRegistry",
    "ActionRequestModel",
    "ActionResponseModel",
    "ReasonModel",
    "Step",
    "SchemaModel",
    "FieldModel",
    "OperableModel",
]



import asyncio
from collections import deque
from typing import Any

from lionabc import BaseManager, Observable
from lionabc.exceptions import ItemNotFoundError, LionValueError

from lion_core.communication.mail import Mail, Package, PackageCategory
from lion_core.generic.exchange import Exchange
from lion_core.generic.pile import Pile, pile
from lion_core.generic.utils import to_list_type
from lion_core.types import IDTypes, LnID


class MailManager(BaseManager):
    """
    Manages mail operations for multiple sources in the Lion framework.

    This class handles the collection, distribution, and management of mail
    between different sources within the system.
    """

    def __init__(
        self, sources: IDTypes.ItemOnly | IDTypes.ItemSeq = None
    ) -> None:
        """
        Initialize a MailManager instance.

        Args:
            sources (List[Any], optional): Initial list of mail sources to
                manage.
        """
        self.sources: Pile[Observable] = pile()
        self.mails: dict[str, dict[str, deque]] = {}
        self.execute_stop: bool = False

        if sources:
            self.add_sources(sources)

    def add_sources(
        self, sources: IDTypes.ItemOnly | IDTypes.ItemSeq, /
    ) -> None:
        """
        Add new sources to the MailManager.

        Args:
            sources (Any): The sources to add. Can be a single source or
                a list of sources.

        Raises:
            LionValueError: If adding the source(s) fails.
        """
        try:
            sources = to_list_type(sources)
            self.sources.include(sources)
            for item in sources:
                self.mails[item.ln_id] = {}
        except Exception as e:
            raise LionValueError("Failed to add source.") from e

    @staticmethod
    def create_mail(
        sender: IDTypes.Ref,
        recipient: IDTypes.Ref,
        category: PackageCategory | str,
        package: Any,
        request_source: Any = None,
    ) -> Mail:
        """
        Create a new Mail object.

        Args:
            sender (str): The ID of the sender.
            recipient (str): The ID of the recipient.
            category (str): The category of the mail.
            package (Any): The content of the package.
            request_source (Any, optional): The source of the request.

        Returns:
            Mail: A new Mail object.
        """
        pack = Package(
            category=category, package=package, request_source=request_source
        )
        return Mail(sender=sender, recipient=recipient, package=pack)

    def delete_source(self, source_id: LnID) -> None:
        """
        Delete a source from the MailManager.

        Args:
            source_id (str): The ID of the source to delete.

        Raises:
            LionValueError: If the source does not exist.
        """
        if source_id not in self.sources:
            raise ItemNotFoundError(f"Source {source_id} does not exist.")
        self.sources.pop(source_id)
        self.mails.pop(source_id)

    def collect(self, sender: LnID) -> None:
        """Collect mail from a specific sender."""
        if sender not in self.sources:
            raise ItemNotFoundError(f"Sender source {sender} does not exist.")
        mailbox: Exchange = (
            self.sources[sender]
            if isinstance(self.sources[sender], Exchange)
            else self.sources[sender].mailbox
        )
        while mailbox.pending_outs.size() > 0:
            mail_id = mailbox.pending_outs.popleft()
            mail: Mail = mailbox.pile_.pop(mail_id)
            if mail.recipient not in self.sources:
                rec_ = mail.recipient
                raise ItemNotFoundError(
                    f"Recipient source {rec_} does not exist"
                )
            if mail.sender not in self.mails[mail.recipient]:
                self.mails[mail.recipient].update({mail.sender: deque()})
            self.mails[mail.recipient][mail.sender].append(mail)

    def send(self, recipient: LnID) -> None:
        """Send mail to a specific recipient."""
        if recipient not in self.sources:
            raise ItemNotFoundError(
                f"Recipient source {recipient} does not exist."
            )
        if not self.mails[recipient]:
            return
        for key in list(self.mails[recipient].keys()):
            pending_mails = self.mails[recipient].pop(key)
            mailbox: Exchange = (
                self.sources[recipient]
                if isinstance(self.sources[recipient], Exchange)
                else self.sources[recipient].mailbox
            )
            while pending_mails:
                mail = pending_mails.popleft()
                mailbox.include(mail, direction="in")

    def collect_all(self) -> None:
        """Collect mail from all sources."""
        for source in self.sources:
            self.collect(sender=source.ln_id)

    def send_all(self) -> None:
        """Send mail to all recipients."""
        for source in self.sources:
            self.send(recipient=source.ln_id)

    async def execute(self, refresh_time: int = 1) -> None:
        """
        Execute mail collection and sending process asynchronously.

        This method runs in a loop, collecting and sending mail at
            regular intervals.

        Args:
            refresh_time (int, optional): The time to wait between
                each cycle in seconds. Defaults to 1.
        """
        while not self.execute_stop:
            self.collect_all()
            self.send_all()
            await asyncio.sleep(refresh_time)


# File: lion_core/communication/mail_manager.py



from lionabc.exceptions import LionValueError

from lion_core.communication import ActionRequest, ActionResponse
from lion_core.protocols import ActionResponseModel


def create_action_response(
    *,
    action_request: ActionRequest,
    action_response_model: ActionResponseModel = None,
    action_response: ActionResponse | None = None,
) -> ActionResponse:
    """Creates or updates an ActionResponse based on the given parameters.

    Args:
        action_request: The corresponding ActionRequest.
        action_response_model: Model containing the response data.
        sender: The sender of the response. Defaults to None.
        action_response: Existing ActionResponse to update.
            Defaults to None.

    Returns:
        ActionResponse: A new or updated ActionResponse instance.

    Raises:
        LionValueError: If action_request is not an ActionRequest or
            if action_response is provided but not an ActionResponse.
    """
    if not isinstance(action_request, ActionRequest):
        raise LionValueError(
            "Error: please provide a corresponding action request for an "
            "action response."
        )

    if action_response:
        if not isinstance(action_response, ActionResponse):
            raise LionValueError(
                "Error: action response must be an instance of ActionResponse."
            )
        if action_request.is_responded:
            raise ValueError("Error: action request already has a response.")
        action_request.content["action_response_id"] = action_response.ln_id
        return action_response

    return ActionResponse(
        action_request=action_request,
        output=action_response_model.output,
    )



from lionfuncs import to_dict

from lion_core.protocols.models import FieldModel

_arguments_description = (
    "Provide the arguments to pass to the function as a "
    "dictionary. **Use "
    "argument names and types as specified in the "
    "`tool_schemas`; do not "
    "invent argument names.**"
)


def validate_arguments(cls, value):
    return to_dict(
        value,
        fuzzy_parse=True,
        suppress=True,
        recursive=True,
    )


ARGUMENTS_FIELD = FieldModel(
    name="arguments",
    annotation=dict,
    default_factory=dict,
    title="Action Arguments",
    description=_arguments_description,
    examples=[{"num1": 1, "num2": 2}, {"x": "hello", "y": "world"}],
    validator=validate_arguments,
    validator_kwargs={"mode": "before"},
)



from lionfuncs import to_num

from lion_core.protocols.models import FieldModel

_description = (
    "Provide an objective numeric confidence score between 0 and 1 (with 3 "
    "decimal places) indicating how likely you successfully achieved the task"
    " according to user expectation. Interpret the score as:\n"
    "- **1**: Very confident in a good job.\n"
    "- **0**: Not confident at all.\n"
    "- **[0.8, 1]**: You can continue the path of reasoning if needed.\n"
    "- **[0.5, 0.8)**: Recheck your reasoning and consider reverting to a "
    "previous, more confident reasoning path.\n"
    "- **[0, 0.5)**: Stop because the reasoning is starting to be off track."
)


def validate_confidence_score(cls, value) -> float:
    try:
        return to_num(
            value,
            upper_bound=1,
            lower_bound=0,
            num_type=float,
            precision=3,
        )
    except Exception:
        return -1


CONFIDENCE_SCORE_FIELD = FieldModel(
    name="confidence_score",
    annotation=float | None,
    default=None,
    title="Confidence Score",
    description=_description,
    examples=[0.821, 0.257, 0.923, 0.439],
    validator=validate_confidence_score,
    validator_kwargs={"mode": "before"},
)

__all__ = ["CONFIDENCE_SCORE_FIELD"]



from typing import Literal

from pydantic import BaseModel, JsonValue

from lion_core.communication import (
    ActionRequest,
    ActionResponse,
    AssistantResponse,
    Instruction,
    System,
)
from lion_core.protocols import ActionRequestModel, ActionResponseModel
from lion_core.types import IDTypes

from .msg_handlers import (
    create_action_request,
    create_action_response,
    create_assistant_response,
    create_instruction,
    create_system,
)


def create_message(
    sender: IDTypes.SenderRecipient = None,
    recipient: IDTypes.SenderRecipient = None,
    instruction: Instruction | JsonValue = None,
    context: JsonValue = None,
    guidance: JsonValue = None,
    plain_content: str = None,
    request_fields: list[str] | dict = None,
    request_model: type[BaseModel] | BaseModel = None,
    system: System | JsonValue = None,
    system_sender: IDTypes.SenderRecipient = None,
    system_datetime: bool | str = None,
    images: list = None,
    image_detail: Literal["low", "high", "auto"] | None = None,
    assistant_response: AssistantResponse | JsonValue = None,
    action_request: ActionRequest = None,
    action_response: ActionResponse = None,
    action_request_model: ActionRequestModel = None,
    action_response_model: ActionResponseModel = None,
):
    """Create a message based on the provided parameters.

    Args:
        sender: The sender of the message.
        recipient: The recipient of the message.
        instruction: Instruction content or object.
        context: Additional context for the message.
        guidance: Guidance information.
        plain_content: Plain text content.
        request_fields: Fields for the request.
        request_model: Model for the request.
        system: System message content or object.
        system_sender: The sender of the system message.
        system_datetime: System datetime information.
        images: List of images.
        image_detail: Image detail level.
        assistant_response: Assistant response content or object.
        action_request: Action request message.
        action_response: Action response message.
        action_request_model: Action request model.
        action_response_model: Action response model.

    Returns:
        A message object of the appropriate type.

    Raises:
        ValueError: If multiple message roles are provided or if an
            action response is provided without an action request.
    """
    """kwargs for additional instruction context"""
    if (
        len(
            [
                i
                for i in [instruction, system, assistant_response]
                if i is not None
            ],
        )
        > 1
    ):
        raise ValueError("Error: Message can only have one role")

    if action_response_model or action_response:
        if not action_request:
            raise ValueError(
                "Error: Action response must have an action request."
            )
        return create_action_response(
            action_request=action_request,
            action_response_model=action_response_model,
            action_response=action_response,
        )
    if action_request_model or action_request:
        return create_action_request(
            action_request_model=action_request_model,
            sender=sender,
            recipient=recipient,
            action_request=action_request,
        )
    if system:
        return create_system(
            system=system,
            sender=system_sender,
            recipient=recipient,
            system_datetime=system_datetime,
        )
    if assistant_response:
        return create_assistant_response(
            sender=sender,
            recipient=recipient,
            assistant_response=assistant_response,
        )
    return create_instruction(
        sender=sender,
        recipient=recipient,
        instruction=instruction,
        context=context,
        guidance=guidance,
        request_fields=request_fields,
        images=images,
        image_detail=image_detail,
        request_model=request_model,
        plain_content=plain_content,
    )



from .create_action_request_msg import create_action_request
from .create_action_response_msg import create_action_response
from .create_assistant_response_msg import create_assistant_response
from .create_instruction_msg import create_instruction
from .create_system_msg import create_system
from .validate_message import validate_message

__all__ = [
    "create_action_request",
    "create_action_response",
    "create_assistant_response",
    "create_instruction",
    "create_system",
    "create_message",
    "validate_message",
]



from collections.abc import Callable
from typing import Any, TypeVar

from lionabc.exceptions import LionValueError
from lionfuncs import unique_hash
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    PrivateAttr,
    field_serializer,
    field_validator,
)
from pydantic.fields import FieldInfo
from pydantic_core import PydanticUndefined

from lion_core.types import LN_UNDEFINED

FIELD_NAME = TypeVar("FIELD_NAME", bound=str)


def clean_dump(obj: BaseModel) -> dict:
    dict_ = obj.model_dump()
    for i in list(dict_.keys()):
        if dict_[i] is LN_UNDEFINED:
            dict_.pop(i)
    return dict_


class ParamsModel(BaseModel):

    model_config = ConfigDict(
        extra="allow",
        populate_by_name=True,
        arbitrary_types_allowed=True,
        validate_default=False,
    )

    @classmethod
    def check_params(cls, **kwargs) -> dict:
        try:
            a = cls.model_validate(kwargs)
            return a.model_dump()
        except Exception as e:
            raise ValueError("Failed to validate parameters") from e


class SchemaModel(BaseModel):

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        extra="forbid",
        use_enum_values=True,
        populate_by_name=True,
        validate_default=True,
    )

    def to_dict(self) -> dict[str, Any]:
        return clean_dump(self)

    @classmethod
    def schema_keys(cls) -> set:
        return set(cls.model_fields.keys())


class FieldModel(BaseModel):

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        extra="allow",
        use_enum_values=True,
        populate_by_name=True,
    )

    _unique_hash: str = PrivateAttr(lambda: unique_hash(32))

    name: str = Field(..., exclude=True)
    annotation: type | Any = Field(LN_UNDEFINED, exclude=True)
    default: Any = LN_UNDEFINED
    default_factory: Any = LN_UNDEFINED
    title: str | Any = LN_UNDEFINED
    description: str | Any = LN_UNDEFINED
    examples: list | Any = LN_UNDEFINED
    validators: list | Any = LN_UNDEFINED
    exclude: bool | Any = LN_UNDEFINED
    validator: Callable | Any = Field(LN_UNDEFINED, exclude=True)
    validator_kwargs: dict | Any = Field(default_factory=dict, exclude=True)

    def to_dict(self) -> dict[str, Any]:
        return clean_dump(self)

    @property
    def field_info(self) -> FieldInfo:
        annotation = (
            self.annotation if self.annotation is not LN_UNDEFINED else Any
        )
        field_obj = Field(**self.to_dict())  # type: ignore
        field_obj.annotation = annotation
        return field_obj

    @property
    def field_validator(self) -> dict[str, Callable]:
        if self.validator is LN_UNDEFINED:
            return None
        kwargs = self.validator_kwargs or {}
        return {
            f"{self.name}_validator": field_validator(self.name, **kwargs)(
                self.validator
            )
        }

    def __hash__(self) -> int:
        return hash(self._unique_hash)


class OperableModel(BaseModel):

    model_config = ConfigDict(
        extra="forbid",
        arbitrary_types_allowed=True,
        use_enum_values=True,
        populate_by_name=True,
    )

    extra_fields: dict[str, Any] = Field(default_factory=dict)

    @field_serializer("extra_fields")
    def _serialize_extra_fields(
        self,
        value: dict[str, FieldInfo],
    ) -> dict[str, Any]:
        """Custom serializer for extra fields."""
        output_dict = {}
        for k in value.keys():
            k_value = self.__dict__.get(k)
            output_dict[k] = k_value
        return output_dict

    @field_validator("extra_fields")
    def _validate_extra_fields(
        cls,
        value: dict[str, dict | FieldInfo | FieldModel],
    ) -> dict[str, FieldInfo]:
        """Custom validator for extra fields."""
        if not isinstance(value, dict):
            raise LionValueError("Extra fields must be a dictionary")

        out_ = {}
        for k, v in value.items():
            v = v.to_dict() if isinstance(v, FieldModel) else v
            out_[k] = Field(**v) if isinstance(v, dict) else v

        return out_

    @property
    def all_fields(self) -> dict[str, FieldInfo]:
        """
        Get all fields including model fields and extra fields.

        Returns:
            dict[str, FieldInfo]: A dictionary containing all fields.
        """
        return {**self.model_fields, **self.extra_fields}

    def add_field(
        self,
        field_name: FIELD_NAME,
        /,
        value: Any = LN_UNDEFINED,
        annotation: type = LN_UNDEFINED,
        field_obj: FieldInfo = LN_UNDEFINED,
        field_model: FieldModel = LN_UNDEFINED,
        **kwargs,
    ) -> None:
        """
        Add a new field to the component's extra fields.

        Args:
            field_name: The name of the field to add.
            value: The value of the field.
            annotation: Type annotation for the field.
            field_obj: A pre-configured FieldInfo object.
            **kwargs: Additional keyword arguments for Field configuration.

        Raises:
            LionValueError: If the field already exists.
        """
        if field_name in self.all_fields:
            raise LionValueError(f"Field '{field_name}' already exists")

        self.update_field(
            field_name,
            value=value,
            annotation=annotation,
            field_obj=field_obj,
            field_model=field_model,
            **kwargs,
        )

    def update_field(
        self,
        field_name: FIELD_NAME,
        /,
        value: Any = LN_UNDEFINED,
        annotation: type = LN_UNDEFINED,
        field_obj: FieldInfo = LN_UNDEFINED,
        field_model: FieldModel = LN_UNDEFINED,
        **kwargs,
    ) -> None:
        """
        Update an existing field or create a new one if it doesn't exist.

        Args:
            field_name: The name of the field to update or create.
            value: The new value for the field.
            annotation: Type annotation for the field.
            field_obj: A pre-configured FieldInfo object.
            **kwargs: Additional keyword arguments for Field configuration.

        Raises:
            ValueError: If both 'default' and 'default_factory' are
                        provided in kwargs.
        """

        # pydanitc Field object cannot have both default and default_factory
        if "default" in kwargs and "default_factory" in kwargs:
            raise ValueError(
                "Cannot provide both 'default' and 'default_factory'",
            )

        if field_obj and field_model:
            raise ValueError(
                "Cannot provide both 'field_obj' and 'field_model'",
            )

        # handle field_obj
        if field_obj is not LN_UNDEFINED:
            if not isinstance(field_obj, FieldInfo):
                raise ValueError(
                    "Invalid field_obj, should be a pydantic FieldInfo object"
                )
            self.extra_fields[field_name] = field_obj

        if field_model is not LN_UNDEFINED:
            if not isinstance(field_model, FieldModel):
                raise ValueError(
                    "Invalid field_model, should be a FieldModel object"
                )
            self.extra_fields[field_name] = Field(**field_model.to_dict())

        # handle kwargs
        if kwargs:
            if field_name in self.all_fields:  # existing field
                for k, v in kwargs.items():
                    self.field_setattr(field_name, k, v)
            else:
                self.extra_fields[field_name] = Field(**kwargs)

        # handle no explicit defined field
        if field_obj is LN_UNDEFINED and not kwargs:
            if field_name not in self.all_fields:
                self.extra_fields[field_name] = Field()

        field_obj = self.all_fields[field_name]

        # handle annotation
        if annotation is not LN_UNDEFINED:
            field_obj.annotation = annotation
        if not field_obj.annotation:
            field_obj.annotation = Any

        # handle value
        if value is LN_UNDEFINED:
            if getattr(self, field_name, LN_UNDEFINED) is not LN_UNDEFINED:
                value = getattr(self, field_name)

            elif getattr(field_obj, "default") is not PydanticUndefined:
                value = field_obj.default

            elif getattr(field_obj, "default_factory"):
                value = field_obj.default_factory()

        setattr(self, field_name, value)

    # field management methods
    def field_setattr(
        self,
        field_name: FIELD_NAME,
        attr: str,
        value: Any,
        /,
    ) -> None:
        """Set the value of a field attribute."""
        all_fields = self.all_fields
        if field_name not in all_fields:
            raise KeyError(f"Field {field_name} not found in object fields.")
        field_obj = all_fields[field_name]
        if hasattr(field_obj, attr):
            setattr(field_obj, attr, value)
        else:
            if not isinstance(field_obj.json_schema_extra, dict):
                field_obj.json_schema_extra = {}
            field_obj.json_schema_extra[attr] = value

    def field_hasattr(
        self,
        field_name: FIELD_NAME,
        attr: str,
        /,
    ) -> bool:
        """Check if a field has a specific attribute."""
        all_fields = self.all_fields
        if field_name not in all_fields:
            raise KeyError(f"Field {field_name} not found in object fields.")
        field_obj = all_fields[field_name]
        if hasattr(field_obj, attr):
            return True
        elif isinstance(field_obj.json_schema_extra, dict):
            if field_name in field_obj.json_schema_extra:
                return True
        else:
            return False

    def field_getattr(
        self,
        field_name: FIELD_NAME,
        attr: str,
        default: Any = LN_UNDEFINED,
        /,
    ) -> Any:
        """Get the value of a field attribute."""
        all_fields = self.all_fields

        if field_name not in all_fields:
            raise KeyError(f"Field {field_name} not found in object fields.")

        if str(attr).strip("s").lower() == "annotation":
            return self.model_fields[field_name].annotation

        field_obj = all_fields[field_name]

        # check fieldinfo attr
        value = getattr(field_obj, attr, LN_UNDEFINED)
        if value is not LN_UNDEFINED:
            return value
        else:
            if isinstance(field_obj.json_schema_extra, dict):
                value = field_obj.json_schema_extra.get(attr, LN_UNDEFINED)
                if value is not LN_UNDEFINED:
                    return value

        # undefined attr
        if default is not LN_UNDEFINED:
            return default
        else:
            raise AttributeError(
                f"field {field_name} has no attribute {attr}",
            )



from typing import Any, Literal

from pydantic import BaseModel

from lion_core.communication import Instruction


def create_instruction(
    sender: Any = None,
    recipient: Any = None,
    instruction: str | dict = None,
    context: Any = None,
    guidance: str | dict = None,
    request_fields: list | dict = None,
    request_model: type[BaseModel] | BaseModel = None,
    plain_content: str = None,
    images: list = None,
    image_detail: Literal["low", "high", "auto"] | None = None,
    **kwargs,
) -> Instruction:
    """Create or update an Instruction.

    Args:
        sender: The sender of the instruction.
        recipient: The recipient of the instruction.
        instruction: Existing Instruction or instruction content.
        context: Additional context for the instruction.
        guidance: Guidance information.
        request_fields: Fields for the request.
        request_model: Model for the request.
        plain_content: Plain text content.
        images: List of images.
        image_detail: Image detail level.
        **kwargs: Additional keyword arguments.

    Returns:
        Instruction: New or updated Instruction instance.
    """
    if isinstance(instruction, Instruction):
        instruction.update(
            context,
            guidance=guidance,
            request_fields=request_fields,
            plain_content=plain_content,
            request_model=request_model,
            images=images,
            image_detail=image_detail,
            **kwargs,
        )
        if sender:
            instruction.sender = sender
        if recipient:
            instruction.recipient = recipient
    else:
        instruction = Instruction(
            sender=sender,
            recipient=recipient,
            instruction=instruction,
            context=context,
            guidance=guidance,
            request_fields=request_fields,
            request_model=request_model,
            plain_content=plain_content,
            images=images,
            image_detail=image_detail,
            **kwargs,
        )
    return instruction



import re

from lionfuncs import to_json, to_list
from pydantic import BaseModel

DEFAULT_SYSTEM = "You are a helpful AI assistant. Let's think step by step."


def parse_action_request(content: str) -> list[dict]:
    """
    return [{"function": "function_name", "arguments": {...}}, ...]
    """
    json_blocks = _prepare_response(content)
    return _parse_action_request(json_blocks)


def _prepare_response(content: str | dict | BaseModel) -> list[dict | str]:

    if isinstance(content, BaseModel):
        return [content.model_dump()]

    elif content and isinstance(content, dict):
        return [content]

    try:
        return to_list(to_json(content, fuzzy_parse=True), flatten=True)
    except Exception:
        pass

    try:
        pattern2 = r"```python\s*(.*?)\s*```"
        json_blocks = re.findall(pattern2, content, re.DOTALL)
        return to_list(to_json(json_blocks, fuzzy_parse=True), flatten=True)
    except Exception:
        return []


def _parse_action_request(json_blocks: list[dict]) -> list[dict]:
    if not json_blocks:
        return []

    out = []
    for i in json_blocks:
        j = {}
        if isinstance(i, dict):
            for k, v in i.items():
                k = (
                    k.replace("action_", "")
                    .replace("recipient_", "")
                    .replace("s", "")
                )
                if k in ["name", "function", "recipient"]:
                    j["function"] = v
                elif k in ["parameter", "argument", "arg"]:
                    j["arguments"] = v
            if list(j.keys()) == ["function", "arguments"]:
                out.append(j)

    return out



from lion_service import iModel
from lionabc import AbstractSpace
from pydantic import field_validator, model_validator

from lion_core.communication import System
from lion_core.communication.utils import validate_sender_recipient
from lion_core.generic.node import Node
from lion_core.session.msg_handlers import create_system
from lion_core.sys_utils import SysUtil
from lion_core.types import IDTypes


class BaseSession(Node, AbstractSpace):
    system: System | None = None
    user: IDTypes.SenderRecipient | None = None
    imodel: iModel | None = None
    name: str | None = None

    @field_validator("user", mode="before")
    def validate_user(cls, value: IDTypes.SenderRecipient):
        v = validate_sender_recipient(value)
        if v in ["system", "assistant"]:
            raise ValueError(f"Invalid user: {v}")
        return v

    @model_validator(mode="before")
    def validate_system(cls, data: dict):
        system = data.pop("system", None)
        if system is None:
            return data

        sender = data.pop("system_sender", None)
        system_datetime = data.pop("system_datetime", None)

        system = create_system(
            system=system,
            sender=sender,
            system_datetime=system_datetime,
        )
        data["system"] = system
        return data

    @model_validator(mode="after")
    def check_system_recipient(self):
        if self.system is None:
            return self
        if not SysUtil.is_id(self.system.recipient):
            self.system.recipient = self.ln_id
        return self


__all__ = ["BaseSession"]
# File: lion_core/session/base.py



from typing import Any

from lion_core.communication import AssistantResponse


def create_assistant_response(
    sender: Any = None,
    recipient: Any = None,
    assistant_response: AssistantResponse | Any = None,
) -> AssistantResponse:
    """Create or return an AssistantResponse.

    Args:
        sender: The sender of the response.
        recipient: The recipient of the response.
        assistant_response: Existing AssistantResponse or response content.

    Returns:
        AssistantResponse: New or existing AssistantResponse instance.
    """
    if isinstance(assistant_response, AssistantResponse):
        if sender:
            assistant_response.sender = sender
        if recipient:
            assistant_response.recipient = recipient
        return assistant_response

    return AssistantResponse(
        assistant_response=assistant_response,
        sender=sender,
        recipient=recipient,
    )



from collections.abc import Mapping, Sequence
from typing import Annotated, Generic, Literal, TypeAlias, TypeVar

from lionabc import Container, Observable, Ordering
from lionfuncs import LN_UNDEFINED, LionUndefinedType
from pydantic import Field

# TypeVar for generic type constraints
T = TypeVar("T", bound=Observable)

# Basic ID type
LnID: TypeAlias = Annotated[str, Field(description="Lion ID string ('ln...')")]


class IDTypes(Generic[T]):
    """
    Generic type definitions for different ID usage contexts.

    TypeVars:
        T: Type parameter constrained to Observable types
    """

    # For functions that accept either ID or item
    Ref: TypeAlias = LnID | T  # type: ignore

    # For functions requiring just the ID
    IDOnly: TypeAlias = LnID

    # For functions requiring Observable object
    ItemOnly = T  # type: ignore

    # For collections
    IDSeq: TypeAlias = Sequence[LnID] | Ordering[LnID]
    ItemSeq: TypeAlias = (  # type: ignore
        Sequence[T] | Mapping[LnID, T] | Container[LnID | T]
    )
    RefSeq: TypeAlias = IDSeq | ItemSeq

    # For system-level interactions
    SenderRecipient: TypeAlias = (  # type: ignore
        LnID | T | Literal["system", "user", "N/A", "assistant"]
    )

    # example: IDTypes[UserClass].Ref


__all__ = ["LnID", "IDTypes", "LN_UNDEFINED", "LionUndefinedType"]



from lion_core.protocols.models import FieldModel
from lion_core.protocols.operatives.reason_model import ReasonModel

_description = "**Provide a concise reason for the decision made.**"

REASON_FIELD = FieldModel(
    name="reason",
    annotation=ReasonModel | None,
    default=None,
    title="Reason",
    description=_description,
)



from pydantic import BaseModel, field_validator

from lion_core.protocols.fields.confidence_score_ import CONFIDENCE_SCORE_FIELD


class ReasonModel(BaseModel):

    title: str | None = None
    content: str | None = None
    confidence_score: float | None = CONFIDENCE_SCORE_FIELD.field_info

    @field_validator(
        "confidence_score", **CONFIDENCE_SCORE_FIELD.validator_kwargs
    )
    def _validate_confidence(cls, v):
        return CONFIDENCE_SCORE_FIELD.validator(cls, v)



from lionfuncs import validate_boolean

from lion_core.protocols.models import FieldModel

_description = (
    "Specify whether the step requires actions to be "
    "performed. If **True**, the actions in `action_requests` "
    "must be performed. If **False**, the actions in "
    "`action_requests` are optional. If no tool_schemas"
    " are provided, this field is ignored."
)


def validate_action_required(cls, value) -> bool:
    try:
        return validate_boolean(value)
    except Exception:
        return False


ACTION_REQUIRED_FIELD = FieldModel(
    name="action_required",
    annotation=bool,
    default=False,
    title="Action Required",
    description=_description,
    validator=validate_action_required,
    validator_kwargs={"mode": "before"},
)



import logging
from collections.abc import Callable
from pathlib import Path
from typing import Any, Literal

import pandas as pd
from lion_service import iModel
from lionabc import Observable, Real, Traversal
from lionfuncs import (
    alcall,
    break_down_pydantic_annotation,
    copy,
    is_same_dtype,
    to_json,
    validate_mapping,
)
from pydantic import BaseModel, ConfigDict, Field, model_validator
from pydantic.fields import FieldInfo
from pydantic.types import JsonValue

from lion_core.action import Tool, ToolManager
from lion_core.action.tool_manager import FUNCTOOL
from lion_core.communication import (
    MESSAGE_FIELDS,
    ActionRequest,
    ActionResponse,
    AssistantResponse,
    Instruction,
    Mail,
    Package,
    RoledMessage,
    System,
)
from lion_core.communication.package import PackageCategory
from lion_core.generic import Exchange, Pile, Progression, progression
from lion_core.log_manager import LogManager
from lion_core.protocols.models import FieldModel
from lion_core.protocols.operatives.action_model import (
    ActionRequestModel,
    ActionResponseModel,
)
from lion_core.protocols.operatives.step import Step
from lion_core.protocols.operatives.utils import PrepFieldParamsModel
from lion_core.session.base import BaseSession
from lion_core.session.msg_handlers import create_instruction, validate_message
from lion_core.setting import Settings
from lion_core.sys_utils import SysUtil
from lion_core.types import LN_UNDEFINED, IDTypes

from .utils import create_message


class Branch(BaseSession, Traversal):
    """
    Represents a branch in the conversation tree with tools and messages.

    This class manages a conversation branch, including messages, tools,
    and communication within the branch.

    Attributes:
        messages (Pile): Collection of messages in the branch.
        tool_manager (ToolManager): Manages tools available in the branch.
        mailbox (Exchange): Handles mail communication for the branch.
        progress (Progression): Tracks the order of messages.
        system (System | None): System message for the branch.
        user (str): Identifier for the user in the conversation.
        imodel (iModel | None): AI model associated with the branch.
        operative_model (type[BaseModel] | None): Model for operation results.
    """

    messages: Pile = Field(default_factory=Pile)
    tool_manager: ToolManager = Field(default_factory=ToolManager)
    mailbox: Exchange = Field(default_factory=Exchange)
    progress: Progression = Field(default_factory=Progression)
    operative_model: type[BaseModel] | None = None
    message_log_manager: LogManager | None = Field(None, exclude=True)
    action_log_manager: LogManager | None = Field(None, exclude=True)
    save_on_clear: bool = Field(default=True)

    @model_validator(mode="before")
    def _validate_input(cls, data: dict) -> dict:
        messages = data.pop("messages", None)
        persist_dir = data.pop("persist_dir", None)
        data["message_log_manager"] = data.get(
            "message_log_manager",
            LogManager(
                persist_dir=persist_dir,
                subfolder="messages",
                file_prefix="message",
                capacity=128,
            ),
        )
        data["action_log_manager"] = data.get(
            "action_log_manager",
            LogManager(
                persist_dir=persist_dir,
                subfolder="actions",
                file_prefix="action",
                capacity=128,
            ),
        )
        data["messages"] = Pile(
            validate_message(messages),
            {RoledMessage},
            strict=False,
        )
        data["progress"] = progression(
            list(data.pop("messages", [])),
        )
        data["tool_manager"] = data.pop(
            "tool_manager",
            ToolManager(),
        )
        data["mailbox"] = data.pop(
            "mailbox",
            Exchange(),
        )
        data["imodel"] = data.pop(
            "imodel", iModel(**Settings.iModel.CHAT.model_dump())
        )
        if "tools" in data:
            tool_manager = data.get("tool_manager", None)
            tool_manager = tool_manager or ToolManager()
            tool_manager.register_tools(data.pop("tools"))
            data["tool_manager"] = tool_manager
        return cls.validate_system(data)

    @model_validator(mode="after")
    def _check_system(self):
        if self.system is None:
            return self
        if self.system not in self.messages:
            self.messages.include(self.system)
            self.progress.insert(0, self.system)
        return self

    def has_messages(self, exclude_system=True) -> bool:
        """
        whether the branch has any messages. If exclude_system is True
        we will ignore system messages in the count.
        """
        if exclude_system:
            return any([not isinstance(i, System) for i in self.messages])
        return self.messages.is_empty()

    def has_logs(self) -> bool:
        if self.action_log_manager.logs or self.message_log_manager.logs:
            return True
        return False

    def set_system(self, system: System) -> None:
        """
        Sets or updates the system message for the branch.
        Do not use this method. use add_message instead.
        We are considering making this method private.

        Args:
            system (System): The new system message to set.
        """
        if len(self.progress) < 1:
            self.messages.include(system)
            self.system = system
            self.progress[0] = self.system
        else:
            self._change_system(system, delete_previous_system=True)
            self.progress[0] = self.system

    async def a_add_message(self, **kwargs):
        """async add a message, check add_message for details"""
        async with self.messages.async_lock:
            self.add_message(**kwargs)

    def add_message(
        self,
        *,
        sender: IDTypes.SenderRecipient = None,
        recipient: IDTypes.SenderRecipient = None,
        instruction: Instruction | JsonValue = None,
        context: JsonValue = None,
        guidance: JsonValue = None,
        plain_content: str = None,
        request_fields: list[str] | dict[str, Any] = None,
        request_model: type[BaseModel] | BaseModel = None,
        system: System | JsonValue = None,
        system_sender: IDTypes.SenderRecipient = None,
        system_datetime: bool | str = None,
        images: list = None,
        image_detail: Literal["low", "high", "auto"] = None,
        assistant_response: AssistantResponse | JsonValue = None,
        action_request: ActionRequest = None,
        action_response: ActionResponse = None,
        action_request_model: ActionRequestModel = None,
        action_response_model: ActionResponseModel = None,
        delete_previous_system: bool = None,
        metadata: dict = None,
    ):
        """Adds a message to the branch. Only use this to add a message"""

        _msg = create_message(
            sender=sender,
            recipient=recipient,
            instruction=instruction,
            context=context,
            guidance=guidance,
            plain_content=plain_content,
            request_fields=request_fields,
            request_model=request_model,
            system=system,
            system_sender=system_sender,
            system_datetime=system_datetime,
            images=images,
            image_detail=image_detail,
            assistant_response=assistant_response,
            action_request=action_request,
            action_response=action_response,
            action_request_model=action_request_model,
            action_response_model=action_response_model,
        )

        if isinstance(_msg, System):
            _msg.recipient = self.ln_id
            _msg.sender = system_sender or "system"
            self._change_system(_msg, delete_previous_system)

        if isinstance(_msg, Instruction):
            _sender = copy(_msg.sender)
            if not _sender or _sender == "N/A":
                _msg.sender = sender or self.user
            _msg.recipient = recipient or self.ln_id

        if isinstance(_msg, AssistantResponse):
            _sender = copy(_msg.sender)
            if not _sender or _sender == "N/A":
                _msg.sender = sender or self.ln_id
            _msg.recipient = recipient or self.user or "user"

        if isinstance(_msg, ActionRequest):
            _sender = copy(_msg.sender)
            if not SysUtil.is_id(_sender):
                _msg.sender = sender or self.ln_id

            _msg.recipient = copy(_msg.recipient)

            if not SysUtil.is_id(_msg.recipient):
                if _msg.function in self.tool_manager.registry:
                    _msg.recipient = self.tool_manager.registry[
                        _msg.function
                    ].ln_id
                else:
                    _msg.recipient = "N/A"

        if isinstance(_msg, ActionResponse):
            _sender = copy(_msg.sender)
            if not SysUtil.is_id(_sender):
                if _msg.function in self.tool_manager.registry:
                    _msg.sender = self.tool_manager.registry[
                        _msg.function
                    ].ln_id
                else:
                    _msg.sender = "N/A"

            _recipient = copy(_msg.recipient)
            if not SysUtil.is_id(_recipient):
                _msg.recipient = recipient or self.ln_id

        if metadata:
            _msg.metadata.update(metadata, ["extra"])

        if _msg in self.messages:
            self.messages[_msg] = _msg
        else:
            self.messages.include(_msg)

        # we are paranoid about losing any data, thus we keep a
        # log of all messages in addition to the messages themselves
        log = _msg.to_log()
        self.message_log_manager.logs.include(log)

    async def aclear_messages(self):
        """async clear messages, check clear_messages for details"""
        async with self.messages.async_lock:
            self.clear_messages()

    def clear_messages(self) -> None:
        """Clears all messages from the branch except the system message."""
        if self.save_on_clear:
            self.dump_log()

        self.messages.clear()
        self.progress.clear()
        if self.system:
            self.messages.include(self.system)
            self.progress.insert(0, self.system)

    def _change_system(
        self,
        system: System,
        delete_previous_system: bool = False,
    ) -> None:
        """
        Change the system message.

        Args:
            system: The new system message.
            delete_previous_system: If True, delete the previous system
                message.
        """
        old_system = self.system
        self.system = system
        self.messages.insert(0, self.system)

        if delete_previous_system:
            self.messages.exclude(old_system)

    def send(
        self,
        recipient: IDTypes.Ref,
        category: PackageCategory | str,
        package: Any,
        request_source: IDTypes.IDOnly,
    ) -> None:
        """
        Sends a mail to a recipient.

        Args:
            recipient (str): The recipient's ID.
            category (str): The category of the mail.
            package (Any): The content of the mail.
            request_source (str): The source of the request.
        """
        package = Package(
            category=category,
            package=package,
            request_source=request_source,
        )

        mail = Mail(
            sender=self.ln_id,
            recipient=recipient,
            package=package,
        )
        self.mailbox.include(mail, "out")

    def receive(
        self,
        sender: IDTypes.Ref,
        message: bool = False,
        tool: bool = False,
        imodel: bool = False,
    ) -> None:
        """
        Receives mail from a sender.

        Args:
            sender (str): The ID of the sender.
            message (bool): Whether to process message mails.
            tool (bool): Whether to process tool mails.
            imodel (bool): Whether to process imodel mails.

        Raises:
            ValueError: If the sender does not exist or the mail category is
            invalid.
        """
        skipped_requests = progression()
        if not isinstance(sender, str):
            sender = sender.ln_id
        if sender not in self.mailbox.pending_ins.keys():
            raise ValueError(f"No package from {sender}")
        while self.mailbox.pending_ins[sender].size() > 0:
            mail_id = self.mailbox.pending_ins[sender].popleft()
            mail: Mail = self.mailbox.pile_[mail_id]

            if mail.category == "message" and message:
                if not isinstance(mail.package.package, RoledMessage):
                    raise ValueError("Invalid message format")
                new_message = mail.package.package.clone()
                new_message.sender = mail.sender
                new_message.recipient = self.ln_id
                self.messages.include(new_message)
                self.mailbox.pile_.pop(mail_id)

            elif mail.category == "tool" and tool:
                if not isinstance(mail.package.package, Tool):
                    raise ValueError("Invalid tools format")
                self.tool_manager.register_tools(mail.package.package)
                self.mailbox.pile_.pop(mail_id)

            elif mail.category == "imodel" and imodel:
                if not isinstance(mail.package.package, iModel):
                    raise ValueError("Invalid iModel format")
                self.imodel = mail.package.package
                self.mailbox.pile_.pop(mail_id)

            else:
                skipped_requests.append(mail)

        self.mailbox.pending_ins[sender] = skipped_requests

        if self.mailbox.pending_ins[sender].size() == 0:
            self.mailbox.pending_ins.pop(sender)

    def receive_all(self) -> None:
        """Receives mail from all senders."""
        for key in list(self.mailbox.pending_ins.keys()):
            self.receive(key)

    @property
    def last_response(self) -> AssistantResponse | None:
        """
        Retrieves the last assistant response in the branch.

        Returns:
            AssistantResponse | None: The last assistant response, if any.
        """
        for i in reversed(self.progress):
            if isinstance(self.messages[i], AssistantResponse):
                return self.messages[i]

    @property
    def assistant_responses(self) -> Pile:
        """
        Retrieves all assistant responses in the branch.

        Returns:
            Pile: A Pile containing all assistant responses.
        """
        return Pile(
            [
                self.messages[i]
                for i in self.progress
                if isinstance(self.messages[i], AssistantResponse)
            ]
        )

    @property
    def action_requests(self) -> Pile:
        """
        Retrieves all action requests in the branch.

        Returns:
            Pile: A Pile containing all action requests.
        """
        return Pile(
            [
                self.messages[i]
                for i in self.progress
                if isinstance(self.messages[i], ActionRequest)
            ]
        )

    @property
    def action_responses(self) -> Pile:
        """
        Retrieves all action responses in the branch.

        Returns:
            Pile: A Pile containing all action responses.
        """
        return Pile(
            [
                self.messages[i]
                for i in self.progress
                if isinstance(self.messages[i], ActionResponse)
            ]
        )

    @property
    def instructions(self) -> Pile:
        """
        Retrieves all instructions in the branch.

        Returns:
            Pile: A Pile containing all instructions.
        """
        return Pile(
            [
                self.messages[i]
                for i in self.progress
                if isinstance(self.messages[i], Instruction)
            ]
        )

    def has_tools(self) -> bool:
        """
        Checks if the branch has any registered tools.

        Returns:
            bool: True if tools are registered, False otherwise.
        """
        return self.tool_manager.registry != {}

    def register_tools(
        self,
        tools: Tool | Callable | list[Callable] | IDTypes[Tool].ItemSeq,
        update: bool = False,
    ) -> None:
        """
        Registers new tools to the tool manager.

        Args:
            tools (Any): Tools to be registered.
        """
        self.tool_manager.register_tools(tools=tools, update=update)

    def delete_tools(
        self,
        tools: Tool | str | list[str] | IDTypes[Tool].RefSeq,
        verbose: bool = True,
    ) -> bool:
        """
        Deletes specified tools from the tool manager.

        Args:
            tools (Any): Tools to be deleted.
            verbose (bool): If True, print status messages.

        Returns:
            bool: True if deletion was successful, False otherwise.
        """
        if not isinstance(tools, list):
            tools = [tools]
        if is_same_dtype(input_=tools, dtype=str):
            for act_ in tools:
                if act_ in self.tool_manager.registry:
                    self.tool_manager.registry.pop(act_)
            if verbose:
                print("tools successfully deleted")
            return True
        elif is_same_dtype(input_=tools, dtype=Tool):
            for act_ in tools:
                if act_.function_name in self.tool_manager.registry:
                    self.tool_manager.registry.pop(act_.function_name)
            if verbose:
                print("tools successfully deleted")
            return True
        if verbose:
            print("tools deletion failed")
        return False

    def to_chat_messages(
        self, progress: IDTypes[RoledMessage].IDSeq = None
    ) -> list[dict[str, Any]]:
        """
        Converts messages to a list of chat message dictionaries.

        Args:
            progress (Progression, optional): Specific progression to convert.

        Returns:
            list[dict[str, Any]]: A list of chat message dictionaries.

        Raises:
            ValueError: If the provided progress is invalid.
        """
        if progress == []:
            return []
        if not all(i in self.messages for i in (progress or self.progress)):
            raise ValueError("Invalid progress")
        return [self.messages[i].chat_msg for i in (progress or self.progress)]

    async def invoke_action(
        self,
        action_request_model: ActionRequestModel = None,
        action_request: ActionRequest = None,
        suppress_errors: bool = False,
    ) -> ActionResponseModel:
        try:
            if action_request and action_request_model:
                raise ValueError(
                    "Cannot have both action_request and action_request_model"
                )
            if not action_request and not action_request_model:
                raise ValueError(
                    "Either action_request or action_request_model must "
                    "be provided"
                )

            if not action_request:
                action_request = ActionRequest(
                    function=action_request_model.function,
                    arguments=action_request_model.arguments,
                    sender=self.ln_id,
                )

            if not action_request_model:
                action_request_model = ActionRequestModel(
                    function=action_request.function,
                    arguments=action_request.arguments,
                )

            if action_request not in self.messages:
                self.add_message(action_request=action_request)
            else:
                self.messages[action_request] = action_request

            result = await self.tool_manager.invoke(
                action_request, log_manager=self.action_log_manager
            )
            action_response_model = ActionResponseModel(
                function=action_request.function,
                arguments=action_request.arguments,
                output=result,
            )
            self.add_message(
                action_request=action_request,
                action_response_model=action_response_model,
            )
            return action_response_model
        except Exception as e:
            if suppress_errors:
                logging.warning(f"Failed to invoke action: {e}")
                return None
            raise e

    def get_tool_schema(
        self, tools: str | Tool | list[Tool | str] | bool
    ) -> dict:
        tools = tools if isinstance(tools, list) else [tools]
        for i in tools:
            if isinstance(i, Tool | Callable) and i not in self.tool_manager:
                self.tool_manager.register_tools(i)
        return self.tool_manager.get_tool_schema(tools)

    async def communicate(
        self,
        instruction: Instruction | JsonValue = None,
        guidance: JsonValue = None,
        context: JsonValue = None,
        sender: IDTypes.SenderRecipient | Observable = None,
        recipient: IDTypes.SenderRecipient | Observable = None,
        progress: Progression | IDTypes.IDSeq = None,
        request_model: type[BaseModel] | BaseModel = None,
        request_fields: dict[str, Any] | list[str] = None,
        imodel: iModel = None,
        images: list = None,
        image_detail: Literal["low", "high", "auto"] = None,
        tools: str | FUNCTOOL | list[FUNCTOOL | str] | bool = None,
        num_parse_retries: int = 0,
        retry_imodel: iModel = None,
        retry_kwargs: dict = {},
        handle_validation: Literal[
            "raise", "return_value", "return_none"
        ] = "return_value",
        skip_validation: bool = False,
        clear_messages: bool = False,
        invoke_action: bool = True,
        **kwargs,
    ):
        imodel = imodel or self.imodel
        retry_imodel = retry_imodel or imodel
        if clear_messages:
            await self.aclear_messages()

        if num_parse_retries > 5:
            logging.warning(
                f"Are you sure you want to retry {num_parse_retries} "
                "times? lowering retry attempts to 5. Suggestion is under 3"
            )
            num_parse_retries = 5

        ins, res = await self._invoke_imodel(
            instruction=instruction,
            guidance=guidance,
            context=context,
            sender=sender,
            recipient=recipient,
            request_model=request_model,
            progress=progress,
            imodel=imodel,
            images=images,
            image_detail=image_detail,
            tool_schemas=self.get_tool_schema(tools) if tools else None,
            **kwargs,
        )
        self.add_message(instruction=ins)
        self.add_message(assistant_response=res)

        action_request_models = None
        action_response_models = None

        if skip_validation:
            return res.response

        if tools:
            action_request_models = ActionRequestModel.create(res.response)

        if action_request_models and invoke_action:
            action_response_models = await alcall(
                action_request_models,
                self.invoke_action,
                suppress_errors=True,
            )

        if action_request_models and not action_response_models:
            for i in action_request_models:
                self.add_message(
                    action_request_model=i,
                    sender=self,
                    recipient=None,
                )

        parse_success = False

        _d = None
        if request_fields or request_model:
            while parse_success is False and num_parse_retries > 0:
                if request_fields:
                    try:
                        _d = to_json(res.response)
                        _d = validate_mapping(
                            _d,
                            request_fields,
                            handle_unmatched="force",
                            fill_value=LN_UNDEFINED,
                        )
                        _d = {k: v for k, v in _d.items() if v != LN_UNDEFINED}
                    except Exception:
                        pass
                    if _d and isinstance(_d, dict):
                        parse_success = True
                        if res not in self.messages:
                            self.add_message(assistant_response=res)

                elif request_model:
                    _d = to_json(res.response)
                    _d = validate_mapping(
                        _d,
                        break_down_pydantic_annotation(request_model),
                        handle_unmatched="force",
                        fill_value=LN_UNDEFINED,
                    )

                    _d = {k: v for k, v in _d.items() if v != LN_UNDEFINED}
                    if _d and isinstance(_d, dict):
                        try:
                            _d = request_model.model_validate(_d)
                            parse_success = True
                            if res not in self.messages:
                                self.add_message(assistant_response=res)
                        except Exception as e:
                            logging.warning(
                                "Failed to parse model response into "
                                f"pydantic model: {e}"
                            )

                if parse_success is False:
                    logging.warning(
                        "Failed to parse response into request "
                        f"format, retrying... with {retry_imodel.model}"
                    )
                    _, res = await self._invoke_imodel(
                        instruction="reformat text into specified model",
                        context=res.response,
                        request_model=request_model,
                        request_fields=request_fields,
                        progress=[],
                        imodel=retry_imodel or imodel,
                        **retry_kwargs,
                    )
                    num_parse_retries -= 1

        if request_fields and not isinstance(_d, dict):
            if handle_validation == "raise":
                raise ValueError(
                    "Failed to parse response into request format"
                )
            if handle_validation == "return_none":
                return None
            if handle_validation == "return_value":
                return res.response

        if request_model and not isinstance(_d, BaseModel):
            if handle_validation == "raise":
                raise ValueError(
                    "Failed to parse response into request format"
                )
            if handle_validation == "return_none":
                return None
            if handle_validation == "return_value":
                return res.response

        return _d if _d else res.response

    async def operate(
        self,
        instruction: Instruction | JsonValue = None,
        guidance: JsonValue = None,
        context: JsonValue = None,
        sender: Real = None,
        recipient: Real = None,
        operative_model: BaseModel | type[BaseModel] = None,
        progress: Progression | list[str] = None,
        imodel: iModel = None,
        reason: bool = True,
        actions: bool = True,
        tools: str | FUNCTOOL | list[FUNCTOOL | str] | bool = None,
        images: list = None,
        image_detail: Literal["low", "high", "auto"] = None,
        num_parse_retries: int = 0,
        retry_imodel: iModel = None,
        retry_kwargs: dict = {},
        clear_messages: bool = False,
        skip_validation: bool = False,
        handle_validation: Literal[
            "raise", "return_value", "return_none"
        ] = "return_value",
        invoke_action: bool = True,
        exclude_fields: list | dict = None,
        include_fields: list | dict = None,
        config_dict: ConfigDict = None,
        frozen: bool = False,
        doc: str = None,
        validators: dict = None,
        use_base_kwargs: bool = False,
        inherit_base: bool = True,
        field_descriptions: dict[str, str] = None,
        extra_fields: dict[str, FieldInfo] = None,
        field_models: list[FieldModel] = None,
        response_kwargs: dict = {},
        **kwargs,
    ) -> BaseModel | str | dict:
        """
        if skip validation,
        action won't be invoked and response will be returned as is

        response_kwargs check Step.parse_request_to_response_model for details
        """

        if clear_messages:
            self.clear_messages()

        if num_parse_retries > 5:
            logging.warning(
                f"Are you sure you want to retry {num_parse_retries} "
                "times? lowering retry attempts to 5. Suggestion is under 3"
            )
            num_parse_retries = 5

        request_params = await self._operate(
            instruction=instruction,
            guidance=guidance,
            context=context,
            sender=sender,
            recipient=recipient,
            operative_model=operative_model,
            progress=progress,
            imodel=imodel,
            reason=reason,
            actions=actions,
            exclude_fields=exclude_fields,
            include_fields=include_fields,
            config_dict=config_dict,
            doc=doc,
            frozen=frozen,
            validators=validators,
            use_base_kwargs=use_base_kwargs,
            inherit_base=inherit_base,
            field_descriptions=field_descriptions,
            extra_fields=extra_fields,
            tool_schemas=self.get_tool_schema(tools) if tools else None,
            images=images,
            image_detail=image_detail,
            num_parse_retries=num_parse_retries,
            retry_imodel=retry_imodel,
            retry_kwargs=retry_kwargs,
            field_models=field_models,
            skip_validation=skip_validation,
            **kwargs,
        )

        request_model = (
            request_params.request_model or request_params.str_dict_response
        )

        if skip_validation:
            return request_model

        if isinstance(request_model, dict | str):
            if handle_validation == "raise":
                raise ValueError(
                    "Operative model validation failed. iModel response"
                    " not parsed into operative model:"
                    f" {request_params.request_model_type.__name__}"
                )
            if handle_validation == "return_none":
                return None
            if handle_validation == "return_value":
                return request_model

        dict_ = request_model.model_dump()
        action_response_models = None
        if (
            actions
            and invoke_action
            and hasattr(request_model, "action_required")
            and request_model.action_required
            and hasattr(request_model, "action_requests")
            and request_model.action_requests
        ):
            action_response_models = await alcall(
                request_model.action_requests,
                self.invoke_action,
                suppress_errors=True,
            )
            action_response_models = [i for i in action_response_models if i]
            dict_["action_responses"] = action_response_models

            try:
                return Step.parse_request_to_response_model(
                    request_params=request_params,
                    data=dict_,
                    **response_kwargs,
                )

            except Exception as e:
                logging.warning(
                    "Failed to parse model response into response"
                    f" operative model area 1: {e}"
                )
                dict_["action_responses"] = action_response_models
                request_model = dict_

        elif (
            hasattr(request_model, "action_requests")
            and request_model.action_requests
        ):
            for i in request_model.action_requests:
                act_req = ActionRequest(
                    function=i.function,
                    arguments=i.arguments,
                    sender=self,
                )
                self.add_message(
                    action_request=act_req,
                    sender=act_req.sender,
                    recipient=None,
                )

        try:
            return Step.parse_request_to_response_model(
                request_params=request_params,
                data=dict_,
                **response_kwargs,
            )

        except Exception as e:
            logging.warning(
                "Failed to parse model response into response "
                f"operative model area 2: {e}"
            )

        return request_model

    async def _invoke_imodel(
        self,
        instruction=None,
        guidance=None,
        context=None,
        sender=None,
        recipient=None,
        request_fields=None,
        request_model: type[BaseModel] = None,
        progress=None,
        imodel: iModel = None,
        tool_schemas=None,
        images: list = None,
        image_detail: Literal["low", "high", "auto"] = None,
        **kwargs,
    ) -> tuple[Instruction, AssistantResponse]:

        ins = create_instruction(
            instruction=instruction,
            guidance=guidance,
            context=context,
            sender=sender or self.user or "user",
            recipient=recipient or self.ln_id,
            request_model=request_model,
            request_fields=request_fields,
            images=images,
            image_detail=image_detail,
        )
        if tool_schemas:
            ins.update_context(tool_schemas=tool_schemas)

        kwargs["messages"] = self.to_chat_messages(progress)
        kwargs["messages"].append(ins.chat_msg)

        imodel = imodel or self.imodel
        api_request = imodel.parse_to_data_model(**kwargs)
        api_response = await imodel.invoke(**api_request)
        res = AssistantResponse(
            assistant_response=api_response,
            sender=self,
            recipient=self.user,
        )
        return ins, res

    async def _operate(
        self,
        instruction=None,
        guidance=None,
        context=None,
        sender=None,
        recipient=None,
        operative_model: type[BaseModel] = None,
        progress=None,
        imodel: iModel = None,
        reason: bool = True,
        actions: bool = True,
        exclude_fields: list | dict | None = None,
        include_fields: list | dict | None = None,
        config_dict: ConfigDict | None = None,
        doc: str | None = None,
        validators: dict = None,
        use_base_kwargs: bool = False,
        inherit_base: bool = True,
        field_descriptions: dict[str, str] | None = None,
        extra_fields: dict[str, FieldInfo] | None = None,
        frozen: bool = False,
        tool_schemas=None,
        images: list = None,
        image_detail: Literal["low", "high", "auto"] = None,
        num_parse_retries: int = 0,
        retry_imodel: iModel = None,
        retry_kwargs: dict = {},
        field_models: list[FieldModel] | None = None,
        skip_validation: bool = False,
        **kwargs,
    ) -> PrepFieldParamsModel | str | dict:
        imodel = imodel or self.imodel
        retry_imodel = retry_imodel or imodel
        request_params: PrepFieldParamsModel = Step.as_request_model_type(
            reason=reason,
            actions=actions,
            exclude_fields=exclude_fields,
            include_fields=include_fields,
            operative_model=operative_model,
            config_dict=config_dict,
            doc=doc,
            frozen=frozen,
            validators=validators,
            use_base_kwargs=use_base_kwargs,
            inherit_base=inherit_base,
            field_descriptions=field_descriptions,
            extra_fields=extra_fields,
            field_models=field_models,
        )
        ins, res = await self._invoke_imodel(
            instruction=instruction,
            guidance=guidance,
            context=context,
            sender=sender,
            recipient=recipient,
            request_model=request_params.request_model_type,
            progress=progress,
            imodel=imodel,
            tool_schemas=tool_schemas,
            images=images,
            image_detail=image_detail,
            **kwargs,
        )

        self.add_message(instruction=ins)
        self.add_message(assistant_response=res)

        request_params.str_dict_response = res.response
        if skip_validation:
            return request_params

        def _validate_pydantic(_s, pyd=True):
            d_ = to_json(_s, fuzzy_parse=True)
            if isinstance(d_, list | tuple) and len(d_) == 1:
                d_ = d_[0]
            if not isinstance(d_, dict):
                raise ValueError(
                    "Failed to parse response into request format"
                )
            d_ = {k: v for k, v in d_.items() if v != LN_UNDEFINED}
            if pyd:
                step_request_model = (
                    request_params.request_model_type.model_validate(d_)
                )
                return d_, step_request_model
            return d_

        step_request_model = None
        dict_ = {}

        try:
            dict_, step_request_model = _validate_pydantic(res.response, True)
            request_params.request_model = step_request_model

        except Exception:
            while num_parse_retries > 0 and step_request_model is None:

                _, res1 = await self._invoke_imodel(
                    instruction="reformat text into specified model",
                    context=res.response,
                    request_model=request_params.request_model_type,
                    progress=[],
                    imodel=retry_imodel,
                    **retry_kwargs,
                )
                try:
                    dict_, step_request_model = _validate_pydantic(
                        res1.response, True
                    )
                    request_params.request_model = step_request_model
                    self.add_message(assistant_response=res1)
                    break

                except Exception:
                    if num_parse_retries == 1:
                        try:
                            dict_ = _validate_pydantic(res1.response, False)
                            break
                        except Exception:
                            dict_ = res.response
                            logging.warning(
                                f"Failed to parse response: {res.response}. "
                                "max retries reached, returning raw response"
                            )
                            break

                    logging.warning(
                        f"Failed to parse response: {res.response}. "
                        f"Retrying with {retry_imodel.model} ..."
                    )
                    step_request_model = None
                    num_parse_retries -= 1

        request_params.str_dict_response = dict_
        return request_params

    def dump_log(self, clear: bool = True, persist_path: str | Path = None):
        self.message_log_manager.dump(clear, persist_path)
        self.action_log_manager.dump(clear, persist_path)

    def to_df(self, *, progress: Progression = None) -> pd.DataFrame:
        if progress is None:
            return self.messages.to_df(columns=MESSAGE_FIELDS)
        msgs = [self.messages[i] for i in progress if i in self.messages]
        p = Pile(items=msgs)
        return p.to_df(columns=MESSAGE_FIELDS)


__all__ = ["Branch"]
# File: lion_core/session/branch.py



from lion_core.protocols.models import FieldModel
from lion_core.protocols.operatives.action_model import ActionResponseModel

ACTION_RESPONSES_FIELD = FieldModel(
    name="action_responses",
    annotation=list[ActionResponseModel],
    default_factory=list,
    title="Actions",
    description="**do not fill**",
)



from pydantic import BaseModel, ConfigDict
from pydantic.fields import FieldInfo

from lion_core.protocols.fields.action_requests_ import ACTION_REQUESTS_FIELD
from lion_core.protocols.fields.action_required_ import ACTION_REQUIRED_FIELD
from lion_core.protocols.fields.action_responses_ import ACTION_RESPONSES_FIELD
from lion_core.protocols.fields.reason_ import REASON_FIELD
from lion_core.protocols.models import FieldModel

from .utils import PrepFieldParamsModel


class Step:

    @staticmethod
    def as_request_model_type(
        reason: bool = False,
        actions: bool = False,
        exclude_fields: set | list = None,
        include_fields: set | list = None,
        operative_model: type[BaseModel] | None = None,
        config_dict: ConfigDict | None = None,
        doc: str | None = None,
        validators=None,
        use_base_kwargs: bool = False,
        inherit_base: bool = True,
        field_descriptions: dict[str, str] | None = None,
        frozen: bool = False,
        extra_fields: dict[str, FieldInfo] | None = None,
        use_all_fields: bool = False,
        field_models: list[FieldModel] | None = None,
    ):

        include_fields = include_fields or []
        field_models = field_models or []

        if not isinstance(include_fields, list):
            include_fields = list(include_fields)

        if reason:
            field_models.append(REASON_FIELD)

        if actions:
            field_models.append(ACTION_REQUIRED_FIELD)
            field_models.append(ACTION_REQUESTS_FIELD)
            field_models.append(ACTION_RESPONSES_FIELD)

        params = PrepFieldParamsModel.as_request_model_type(
            exclude_fields=exclude_fields,
            include_fields=include_fields,
            base=operative_model,
            config_dict=config_dict,
            doc=doc,
            validators=validators,
            use_base_kwargs=use_base_kwargs,
            inherit_base=inherit_base,
            field_descriptions=field_descriptions,
            frozen=frozen,
            extra_fields=extra_fields,
            use_all_fields=use_all_fields,
            field_models=field_models,
        )
        return params

    @staticmethod
    def as_response_model_type(
        request_params: PrepFieldParamsModel,
        return_fields: dict | list = None,
        response_validators: dict = None,
        frozen_reponse: bool = False,
        response_config_dict=None,
        response_doc=None,
    ) -> PrepFieldParamsModel:

        field_models = []
        if request_params.request_model.action_required:
            field_models.append(ACTION_RESPONSES_FIELD)

        params = PrepFieldParamsModel.as_response_model_type(
            request_params=request_params,
            field_models=field_models,
            return_fields=return_fields,
            response_validators=response_validators,
            frozen=frozen_reponse,
            response_config_dict=response_config_dict,
            response_doc=response_doc,
        )
        return params

    @staticmethod
    def parse_request_to_response_model(
        request_params: PrepFieldParamsModel,
        data: dict,
        return_fields: dict | list = None,
        response_validators: dict = None,
        frozen_reponse: bool = False,
        response_config_dict=None,
        response_doc=None,
    ) -> BaseModel:
        params = Step.as_response_model_type(
            request_params=request_params,
            return_fields=return_fields,
            response_validators=response_validators,
            frozen_reponse=frozen_reponse,
            response_config_dict=response_config_dict,
            response_doc=response_doc,
        )
        return params.response_model_type.model_validate(data)



from lion_core.protocols.models import FieldModel
from lion_core.protocols.operatives.action_model import ActionRequestModel

_description = (
    "List of actions to be performed if `action_required` "
    "is **True**. Leave empty if no action is required. "
    "**When providing actions, you must choose from the "
    "provided `tool_schemas`. Do not invent function or "
    "argument names.**"
)

ACTION_REQUESTS_FIELD = FieldModel(
    name="action_requests",
    annotation=list[ActionRequestModel],
    default_factory=list,
    title="Actions",
    description=_description,
)



from typing import Any

from lion_core.communication.system import System

from .utils import DEFAULT_SYSTEM


def create_system(
    system: Any = None,
    sender: Any = None,
    recipient: Any = None,
    system_datetime: bool | str = None,
) -> System:
    """Create or update a System message.

    Args:
        system: Existing System or system content.
        sender: The sender of the system message.
        recipient: The recipient of the system message.
        system_datetime: System datetime information.
            ignored if system is a System instance.

    Returns:
        System: New or updated System instance.
    """
    system = system or DEFAULT_SYSTEM

    if isinstance(system, System):
        if sender:
            system.sender = sender
        if recipient:
            system.recipient = recipient
        return system

    return System(
        system=system,
        sender=sender,
        recipient=recipient,
        system_datetime=system_datetime,
    )



"""lion-core."""

import logging

from .log_manager import LogManager
from .setting import Settings
from .version import __version__

__all__ = [
    "Settings",
    "LogManager",
    "__version__",
]

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)



from typing import Any

from lionabc.exceptions import LionValueError

from lion_core.communication.action_request import ActionRequest
from lion_core.protocols import ActionRequestModel


def create_action_request(
    *,
    action_request_model: ActionRequestModel | dict,
    sender: Any = None,
    recipient: Any = None,
    action_request: ActionRequest | None = None,
) -> ActionRequest:
    """Creates or returns an ActionRequest instance.

    Args:
        action_request_model: Model with function and arguments.
        sender: Request sender.
        recipient: Request recipient.
        action_request: Existing ActionRequest, if any.

    Returns:
        ActionRequest: New or existing instance.

    Raises:
        LionValueError: If action_request is invalid.
        ValueError: If sender or recipient is None.
    """
    if action_request:
        if not isinstance(action_request, ActionRequest):
            raise LionValueError(
                "Error: action request must be an instance of ActionRequest."
            )
        if sender:
            action_request.sender = sender
        if recipient:
            action_request.recipient = recipient
        return action_request

    function = (
        action_request_model["function"]
        if isinstance(action_request_model, dict)
        else action_request_model.function
    )
    arguments = (
        action_request_model["arguments"]
        if isinstance(action_request_model, dict)
        else action_request_model.arguments
    )

    return ActionRequest(
        function=function,
        arguments=arguments,
        sender=sender,
        recipient=recipient,
    )
